#include <linux/i2c-dev.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
# include	<stdio.h>
# include	<stdlib.h>
# include	<unistd.h>
# include	<syslog.h>
# include	<string.h>
# include	<mosquitto.h>
# include	"glib.h"
# include	"Sensor.h"

/*
 * This code is identical to the Si7021 code
 * The Si7021 has a few more features, but for the
 * purpose of this program, they have an identical interface
 */
extern int debug;

libmosq_EXPORT int mosquitto_reconnect_delay_set(
    struct 	mosquitto 	*	mosq,
    unsigned 	int 		reconnect_delay,
    unsigned 	int 		reconnect_delay_max,
    bool 		reconnect_exponential_backoff
    )
{
}

const char *
mosquitto_strerror(int errcode)
{
    static char s[32];

    switch (errcode)
    {
    case MOSQ_ERR_SUCCESS: return "MOSQ_ERR_SUCCESS";
    case MOSQ_ERR_NOMEM: return "MOSQ_ERR_NOMEM";
    case MOSQ_ERR_PROTOCOL: return "MOSQ_ERR_PROTOCOL";
    case MOSQ_ERR_INVAL: return "MOSQ_ERR_INVAL";
    case MOSQ_ERR_NO_CONN: return "MOSQ_ERR_NO_CONN";
    case MOSQ_ERR_CONN_REFUSED: return "MOSQ_ERR_CONN_REFUSED";
    case MOSQ_ERR_NOT_FOUND: return "MOSQ_ERR_NOT_FOUND";
    case MOSQ_ERR_CONN_LOST: return "MOSQ_ERR_CONN_LOST";
# if defined(MOSQ_ERR_SSL) /* { */
    case MOSQ_ERR_SSL: return "MOSQ_ERR_SSL";
# endif /* } */
    case MOSQ_ERR_PAYLOAD_SIZE: return "MOSQ_ERR_PAYLOAD_SIZE";
    case MOSQ_ERR_NOT_SUPPORTED: return "MOSQ_ERR_NOT_SUPPORTED";
    case MOSQ_ERR_AUTH: return "MOSQ_ERR_AUTH";
    case MOSQ_ERR_ACL_DENIED: return "MOSQ_ERR_ACL_DENIED";
    case MOSQ_ERR_UNKNOWN: return "MOSQ_ERR_UNKNOWN";
    case MOSQ_ERR_ERRNO: return "MOSQ_ERR_ERRNO";
    default: snprintf(s, sizeof(s), "MOSQ_ERR_?? (%d)", errcode); return s;
    }
}

int
open_device(char *params)
{
    char i2c_dev_name[32];		// Name of the port we will be using
    int i2c;
    int bus, address;
    char *next, *params1;

    bus = strtol(params, &next, 0);
    if (next == params || *next != ':')
    {
        syslog(LOG_ERR, "bad device parameters %s", params);
	return -1;
    }
    params1 = next + 1;	// skip ':'
    address = strtol(params1, &next, 0);
    if (next == params || *next != '\0')
    {
        syslog(LOG_ERR, "bad device parameters %s", params);
	return -1;
    }

    snprintf(i2c_dev_name, sizeof(i2c_dev_name), "/dev/i2c-%u", bus);
    if (access(i2c_dev_name, F_OK) == -1
     || access(i2c_dev_name, R_OK | W_OK) == -1)
    {
	syslog(LOG_ERR, "%s: %m", i2c_dev_name);
	exit(1);
    }

    if (debug)
	syslog(LOG_DEBUG, "opening %s", i2c_dev_name);
    if ((i2c = open(i2c_dev_name, O_RDWR)) == -1)	// Open port for reading and writing
    {
	syslog(LOG_ERR, "Failed to open \"%s\": %m", i2c_dev_name);
	return -1;
    }

    if (debug)
	syslog(LOG_DEBUG, "open OK, selecting slave %#x", address);
    if (ioctl(i2c, I2C_SLAVE, address) == -1)	// Set the the address of the device we wish to speak to
    {
	syslog(LOG_ERR, "Failed to ioctl(i2c, I2C_SLAVE, %#x): %m", address);
	close(i2c);
	return -1;
    }

    return i2c;
}

/*
 * NAME: make_measurement
 * PURPOSE: To measure temperature and humidity with a HTU21D
 * ARGUMENTS: i2c: open file descriptor to I2C driver
 * RETURNS: Nothing
 */
extern GList *
make_measurement(int i2c)
{
    unsigned char buf[3];
    unsigned int failcount;
    ssize_t nwritten, nread;
    unsigned int Temp_Code, RH_Code;
    float temp, hum;
    struct measurement *m;
    GList *results = NULL;
    char s[128];	// temp string
    unsigned int errcnt;

    if (debug)
	syslog(LOG_DEBUG, "measure temperature, No Hold Master Mode");

    buf[0] = 0xFE;	/* Soft reset */
    (void) write(i2c, &buf[1], 1);	/* Just send, don't check anything */

    buf[0] = 0xF3;	// Measure Temperature, No Hold Master Mode */
    errcnt = 0;

    while (errcnt < 10
       && (nwritten = write(i2c, buf, 1)) != 1)	// Send command
    {
	if (nwritten == -1)
	{
	    if (errcnt == 0)
		syslog(LOG_ERR, "Failed to send measurement command %#x: %m", buf[0]);
	    syslog(LOG_NOTICE, "Sending soft reset");
	    buf[1] = 0xFE;	/* Soft reset */
	    (void) write(i2c, &buf[1], 1);

	    errcnt++;
	    usleep(100000);	/* wait 100ms */
	}
    }
    if (errcnt >= 10)
        return NULL;

    if (debug)
        syslog(LOG_DEBUG, "read temperature measurement");
    failcount = 0;
    while ((nread = read(i2c, buf, 3)) != 3)		// Read measurement data
    {
	if (nread == -1 && ++failcount > 2)
	{
	    syslog(LOG_ERR, "Failed to read temperature data: %m");
	    break;
	}
	usleep(100*1000);
	continue;	// Retry
    }
    if (failcount > 2)
	return NULL;

    if (debug)
	syslog(LOG_DEBUG, "raw temperature data=[%02x %02x %02x]", buf[0], buf[1], buf[2]);

    /*
     * Determine temperature from response data
     */
    Temp_Code = (buf[0] << 8) | (buf[1] & ~0x03);	/* mask off low two bits */
    temp = ((175.72 * Temp_Code) / 65536) - 46.85;

    if (debug)
        syslog(LOG_DEBUG, "measure Relative Humidity, No Hold Master Mode");
    buf[0] = 0xF5;	/* Measure Relative Humidity, No Hold Master Mode */
    if ((nwritten = write(i2c, buf, 1)) != 1)	// Send command
    {
	if (nwritten == -1)
	    syslog(LOG_ERR, "Failed to send measurement command %#x: %m", buf[0]);

	return NULL;
    }

    if (debug)
        syslog(LOG_DEBUG, "read Relative Humidity measurement");
    failcount = 0;
    while ((nread = read(i2c, buf, 3)) != 3)		// Read measurement data
    {
	if (nread == -1 && ++failcount > 2)
	    syslog(LOG_ERR, "Failed to read rel. humidity data: %m");
	usleep(25*1000);	// Give device 25ms time for measurement
	continue;	// Retry
    }
    if (failcount > 2)
	return NULL;

    if (debug)
	syslog(LOG_DEBUG, "raw rel. humidity data=[%02x %02x %02x]", buf[0], buf[1], buf[2]);

    /*
     * Determine rel. humidity from response data
     */
    RH_Code = (buf[0] << 8) | (buf[1] & ~0x03);	/* mask off low two bits */
    hum = ((125.0 * RH_Code) / 65536.0) - 6.0;

    /*
     * TEMPERATURE COEFFICIENT COMPENSATION
     */
    hum += -0.15 * (25.0 - temp);

    if (debug)
	syslog(LOG_DEBUG, "temp=%.2f hum=%.2f\n", temp, hum);

    /* Store values */

    m = malloc(sizeof(struct measurement));
    snprintf(s, sizeof(s), "temperatur");
    m->topic = strdup(s);
    snprintf(s, sizeof(s), "%.2f", temp);
    m->value = strdup(s);
    results = g_list_append(results, m);

    m = malloc(sizeof(struct measurement));
    snprintf(s, sizeof(s), "luftfeuchte");
    m->topic = strdup(s);
    snprintf(s, sizeof(s), "%.2f", hum);
    m->value = strdup(s);
    results = g_list_append(results, m);

    return results;
}

char *
default_params()
{
    return "1:0x40";	// bus 1 address 0x40
}
