#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <mosquitto.h>
#include <syslog.h>
#include <alloca.h>
#include <time.h>
# include	"glib.h"
# include	"Sensor.h"

/*
 * Changelog:
 * 20181124: If make_measurement() returns NULL, don't use this result!
 */
# define	DEFAULT_BUS	1
# define	DEFAULT_BROKER	"bananapi:8883"
# define	DEFAULT_DELAY	60	// publish data once a minute */
# define	MAX_TRIES	10	// max. number of attempts to open /dev/i2c...

struct option options[] = {
    { "broker", required_argument, NULL, 'B' },
    { "debug", no_argument, NULL, 'D' },
    { "password", required_argument, NULL, 'P' },
    { "username", required_argument, NULL, 'U' },

    { "params", required_argument, NULL, 'p' },
    { "foreground", no_argument, NULL, 'f' },
    { "delay", required_argument, NULL, 'd' },
    { "help", no_argument, NULL, 'h' },

    { "one-shot", no_argument, NULL, 0x100 },
    { "no-publish", no_argument, NULL, 0x101 },
    { "record-interval", required_argument, NULL, 0x102 },

    { NULL, 0, NULL, '\0' }
};
int outstanding;
int debug;

struct info {
    struct mosquitto *mosq;
    char *prefix;
};
extern void load_config(const char *pathname);
extern int parse_host_port(char *s, char **hostp, int *portp);
extern void get_host_port_from_file(char *path, char **hostp, int *portp);
extern char *find_config(char *name);
extern void create_pid_file(char *pname);
extern char *default_params();
extern int open_device(char *params);
extern GList *make_measurement(int i2c);

void usage(int argc, char *argv[]);
// void my_log_callback(struct mosquitto *mosq, void *userdata, int level, const char *str);
# if LIBMOSQUITTO_VERSION_NUMBER >= 1000000 /* { */
void my_publish_callback(struct mosquitto *mosq, void *obj, int mid);
# else /* }{ */
void my_publish_callback(/* struct mosquitto *mosq, */ void *obj, uint16_t mid);
# endif /* } */
char *mosq_strerror(int res);
# if LIBMOSQUITTO_VERSION_NUMBER < 1000000 /* { */
// For the time being ...
# define	mosquitto_strerror(mosq_errno)	mosq_strerror(mosq_errno)
# endif /* } */
void do_publish(gpointer data, gpointer user_data);
void free_one_measurement(gpointer data);
void show_one_measurement(gpointer data, gpointer user_data);
static int parse_interval(char *s, unsigned int *intvlp);
static void record_measurement(GList *results, char *pname);


int main(int argc, char **argv)
{
    char *prefix;
    int error;
    char *next;
    int ch;
    char *params;	// device parameters
    char *host;
    int port;
    int foreground, reported;
    char *username, *password;
    int delay;
    int one_shot, no_publish;
    char *s;
    char *pname;
    struct mosquitto *mosq;
    int res;
    int seconds;
    unsigned int record_interval = 0;

    pname = strrchr(argv[0], '/');
    if (pname == NULL)
	pname = argv[0];
    else
	pname++;	// Skip '/'

    error = 0;
    openlog(pname, LOG_PERROR | LOG_PID, LOG_DAEMON);

    /* Defaults */
    foreground = 0;
    params = default_params();
    parse_host_port(DEFAULT_BROKER, &host, &port);
    // get_host_port_from_file("/etc/mosquitto/broker", &host, &port);
    load_config("/etc/mosquitto/broker");
    if ((s = find_config("HOST")) != NULL)
	host = s;
    if ((s = find_config("PORT")) != NULL)
	port = atoi(s);
    username = find_config("USER");
    password = find_config("PASS");
    delay = DEFAULT_DELAY;
    one_shot = 0;	/* Continuous operation */
    no_publish = 0;	/* do publish */
    debug = 0;

    while ((ch = getopt_long(argc, argv, "B:DP:U:p:d:fh", options, NULL)) != -1)
	switch (ch)
	{
	case 'B':
	    if (!parse_host_port(optarg, &host, &port))
	    {
	        syslog(LOG_ERR, "%s: bad broker specification: \"%s\"", argv[0], optarg);
		error = 1;
	    }
	    break;
	case 'D': debug = 1;
	    break;
	case 'P': password = optarg;
	    break;
	case 'U': username = optarg;
	    break;
	case 'p': params = optarg;
	    break;
	case 'd': delay = strtol(optarg, &next, 0);
	    if (next == optarg || *next != '\0')
	    {
	        syslog(LOG_ERR, "%s: bad delay \"%s\"", argv[0], optarg);
		error = 1;
	    }
	    break;
	case 'f': foreground = 1;
	    break;
	case 'h': usage(argc, argv);
	    exit(0);
	case 0x100: /* --one-shot */
	    one_shot = 1;
	    break;
	case 0x101: /* --no-publish */
	    no_publish = 1;
	    break;
	case 0x102: /* --record-interval=xxx */
	    if (!parse_interval(optarg, &record_interval))
	    {
	        syslog(LOG_ERR, "%s: bad record-interval \"%s\"", argv[0], optarg);
		error = 1;
	    }
	    break;
	}

    switch (argc - optind)
    {
    case 0: syslog(LOG_NOTICE, "No location given, won't publish");
        no_publish = 1;
	break;
    case 1:
        break;
    default: usage(argc, argv); error = 1; break;
    }

    if (!no_publish)
    {
	if (username == NULL)
	{
	    syslog(LOG_ERR, "no username");
	    error = 1;
	}
	if (password == NULL)
	{
	    syslog(LOG_ERR, "no password");
	    error = 1;
	}
    }

    if (error)
	exit(255);

    if (no_publish)
        prefix = NULL;
    else
    {
	/* Make sure the prefix starts with a slash */
	if (argv[optind][0] == '/')
	    prefix = strdup(argv[optind]);
	else if ((prefix = malloc(1 + strlen(argv[optind]) + 1)) != NULL)
	{
	    prefix[0] = '/';
	    strcpy(&prefix[1], argv[optind]);
	}
	else
	{
	    syslog(LOG_ERR, "malloc failed");
	    exit(2);
	}
    }

    if (!foreground && !one_shot)
    {
	daemon(0, 0);
	create_pid_file(pname);
    }

    syslog(LOG_NOTICE, "%s running ...", pname);

    if (no_publish)
        mosq = NULL;
    else
    {
	char client_id[128];

	mosquitto_lib_init();

	snprintf(client_id, sizeof(client_id), "%s-%u-%u", pname, getpid(), (unsigned long) time(NULL));

	if (debug)
	    syslog(LOG_NOTICE, "client_id=\"%s\"", client_id);

# if LIBMOSQUITTO_VERSION_NUMBER >= 1000000 /* { */
	mosq = mosquitto_new(client_id, true, NULL);
# else /* }{ */
	mosq = mosquitto_new(client_id, NULL);
# endif /* } */

	if (mosq == NULL)
	{
	    syslog(LOG_ERR, "mosquitto_new failed");

	    free(prefix);

	    exit(1);
	}
	// mosquitto_log_callback_set(mosq, my_log_callback);
	mosquitto_publish_callback_set(mosq, my_publish_callback);
	if ((res = mosquitto_username_pw_set(mosq, username, password)))
	{
	    syslog(LOG_ERR, "mosquitto_username_pw_set returns %d", res);
	    mosquitto_lib_cleanup();

	    free(prefix);

	    exit(1);
	}

# if LIBMOSQUITTO_VERSION_NUMBER >= 1000000 /* { */
	mosquitto_reconnect_delay_set(mosq, 5, 60, 1);
	while ((res = mosquitto_connect(mosq, host, port, 60)) != MOSQ_ERR_SUCCESS)
# else /* }{ */
	while ((res = mosquitto_connect(mosq, host, port, 60, 0)) != 0)
# endif /* } */
	{
	    if (!reported)
	    {
		syslog(LOG_ERR, "Connect to host %s, port %d failed: %s", host, port, mosquitto_strerror(res));
		reported = 1;
	    }

	    if (res == MOSQ_ERR_INVAL)	// if the input parameters were invalid ...
	    {
		mosquitto_lib_cleanup();

		free(prefix);

		exit(1);
	    }

	    sleep(seconds);
	    if (seconds < 10)
		seconds++;
	}
	if (debug)
	    syslog(LOG_DEBUG, "mosquitto_connect OK");
    }

    seconds = 1;
    reported = 0;

    while (1)
    {
	int i2c;			// File descriptor
	GList *results;

	if ((i2c = open_device(params)) == -1)
	{
	    syslog(LOG_ERR, "Failed to open i2c-device: %m");
	    sleep(1);
	    continue;   // Retry
	}

	results = make_measurement(i2c);

	close(i2c);

	if (results == NULL)
	    syslog(LOG_DEBUG, "failed to make measurement");
	else if (no_publish)
	{
	    g_list_foreach(results, show_one_measurement, NULL);
	    if (debug)
		syslog(LOG_DEBUG, "won't publish");
	}
	else
	{
	    char temp_s[64], hum_s[64];
	    struct info i;
	    outstanding = 0;

	    if ((record_interval != 0) && ((time(NULL) % record_interval) == 0))
	        record_measurement(results, pname);

	    i.mosq = mosq;
	    i.prefix = prefix;
	    g_list_foreach(results, do_publish, &i);

	    /*
	     * wait for data to be sent
	     */
	    while (outstanding > 0)
	    {
		int rc;

# if LIBMOSQUITTO_VERSION_NUMBER >= 1000000 /* { */
		if ((rc = mosquitto_loop(mosq, -1, 1)) != MOSQ_ERR_SUCCESS)
# else /* }{ */
		if ((rc = mosquitto_loop(mosq, -1)) != MOSQ_ERR_SUCCESS)
# endif /* } */
		{
		    syslog(LOG_ERR, "mosquitto_loop() returns %s", mosquitto_strerror(rc));
		    outstanding = 0;
		}
		if (rc == MOSQ_ERR_NO_CONN || rc == MOSQ_ERR_CONN_LOST || rc == MOSQ_ERR_ERRNO)
		{
		    if ((rc = mosquitto_reconnect(mosq)) != MOSQ_ERR_SUCCESS)
			syslog(LOG_ERR, "mosquitto_reconnect() returns %s", mosquitto_strerror(rc));
		    else
			syslog(LOG_NOTICE, "mosquitto_reconnect() successful");
		}
	    }
	}

	if (results != NULL)
	    g_list_free_full(results, free_one_measurement);

	if (one_shot)
	    break;

	if (delay > 1)
	    sleep(delay - 1);
    }

    mosquitto_disconnect(mosq);

    mosquitto_destroy(mosq);

    free(prefix);

    exit(0);
}

void
usage(int argc, char *argv[])
{
    fprintf(stderr, "usage: %s [-p device-parameters] [-B broker] [-h]\n", argv[0]);
    fprintf(stderr, "\tdefaults: device-parameters=%s, broker=%s\n", default_params(), DEFAULT_BROKER);
    return;
}

# if 0 /* { */
/*
 * NAME: my_log_callback
 * PURPOSE: Callback function for libmosquitto
 * ARGUMENTS: mosq: a valid mosquitto instance
 *      userdata: whatever was passed in the 3rd parameter when mosquitto_new was called
 *      level: the log message level
 *      str: the message string
 * RETURNS: Nothing.
 */
void
my_log_callback(struct mosquitto *mosq, void *userdata, int level, const char *str)
{
    /* Print all log messages regardless of level. */
    switch (level)
    {
    case MOSQ_LOG_INFO: level = LOG_INFO; break;
    case MOSQ_LOG_NOTICE: level = LOG_NOTICE; break;
    case MOSQ_LOG_WARNING: level = LOG_WARNING; break;
    case MOSQ_LOG_ERR: level = LOG_ERR; break;
    case MOSQ_LOG_DEBUG: level = LOG_DEBUG; break;
    }
    syslog(level, "%s", str);
}
# endif /* } */

/*
 * NAME: my_publish_callback
 * PURPOSE: Called when a message has been sent
 * ARGUMENTS: mosq: the mosquitto instance making the callback.
 *      obj: the user data provided in mosquitto_new
 *      mid: the message id of the sent message.
 * RETURNS: Nothing
 */
void
# if LIBMOSQUITTO_VERSION_NUMBER >= 1000000 /* { */
my_publish_callback(struct mosquitto *mosq, void *obj, int mid)
# else /* }{ */
my_publish_callback(/* struct mosquitto *mosq, */ void *obj, uint16_t mid)
# endif /* } */
{
    if (outstanding)
	outstanding--;
    return;
}

# if LIBMOSQUITTO_VERSION_NUMBER < 1000000 /* { */
char *
mosq_strerror(int res)
{
    static char errstr[128];

    switch (res)
    {
    case MOSQ_ERR_SUCCESS: return "MOSQ_ERR_SUCCESS";
    case MOSQ_ERR_NOMEM: return "MOSQ_ERR_NOMEM";
    case MOSQ_ERR_PROTOCOL: return "MOSQ_ERR_PROTOCOL";
    case MOSQ_ERR_INVAL: return "MOSQ_ERR_INVAL";
    case MOSQ_ERR_NO_CONN: return "MOSQ_ERR_NO_CONN";
    case MOSQ_ERR_CONN_REFUSED: return "MOSQ_ERR_CONN_REFUSED";
    case MOSQ_ERR_NOT_FOUND: return "MOSQ_ERR_NOT_FOUND";
    case MOSQ_ERR_CONN_LOST: return "MOSQ_ERR_CONN_LOST";
# if LIBMOSQUITTO_VERSION_NUMBER < 1000000 /* { */
    case MOSQ_ERR_SSL: return "MOSQ_ERR_SSL";
# endif /* } */
    case MOSQ_ERR_PAYLOAD_SIZE: return "MOSQ_ERR_PAYLOAD_SIZE";
    case MOSQ_ERR_NOT_SUPPORTED: return "MOSQ_ERR_NOT_SUPPORTED";
    case MOSQ_ERR_AUTH: return "MOSQ_ERR_AUTH";
    case MOSQ_ERR_ACL_DENIED: return "MOSQ_ERR_ACL_DENIED";
    case MOSQ_ERR_UNKNOWN: return "MOSQ_ERR_UNKNOWN";
    case MOSQ_ERR_ERRNO: return "MOSQ_ERR_ERRNO";
    default: snprintf(errstr, sizeof(errstr), "MOSQ_ERR %d", res); return errstr;
    }
}
# endif /* } */

void
do_publish(gpointer data, gpointer user_data)
{
    struct measurement *m = (struct measurement *)data;
    struct info *info = (struct info *) user_data;
    struct mosquitto *mosq = info->mosq;
    int mosq_err;
    char *topic;
    unsigned int errcnt;

    if (debug)
	syslog(LOG_NOTICE, "%s/%s = %s", info->prefix, m->topic, m->value);

    topic = alloca(strlen(info->prefix) + 1 + strlen(m->topic) + 1);
    sprintf(topic, "%s/%s", info->prefix, m->topic);
    errcnt = 0;

    while (errcnt < 10
       && (mosq_err = mosquitto_publish(mosq, NULL, topic, strlen(m->value), m->value, 1, 1)) != MOSQ_ERR_SUCCESS)
    {
        syslog(LOG_ERR, "mosquitto_publish: %s (%d) reconnecting ...", mosquitto_strerror(mosq_err), mosq_err);

	if ((mosq_err = mosquitto_reconnect(mosq)) != MOSQ_ERR_SUCCESS)
	    syslog(LOG_ERR, "mosquitto_reconnect: %s (%d)", mosquitto_strerror(mosq_err), mosq_err);

	errcnt++;
    }

    if (errcnt < 10)
	outstanding++;

    return;
}

void
show_one_measurement(gpointer data, gpointer user_data)
{
    struct measurement *m = (struct measurement *)data;

    fprintf(stderr, "%s => %s\n", m->topic, m->value);
}

void
free_one_measurement(gpointer data)
{
    struct measurement *m = (struct measurement *)data;

    if (m->topic)
        free(m->topic);
    if (m->value)
        free(m->value);

    return;
}

static int
parse_interval(char *s, unsigned int *intvlp)
{
    char *next;

    if (intvlp == NULL)
        return 0;

    *intvlp = strtoul(s, &next, 10);
    if (strcmp(next, "m") == 0)		/* minutes */
        *intvlp *= 60;
    else if (strcmp(next, "h") == 0)	/* hours */
        *intvlp *= 2600;
    else
        return 0;

    return 1;
}

/*
 * We need to transfer both the current time and the destination file
 * into g_list_foreach
 */
struct log_data {
    struct tm *tv;
    FILE *dst;
    char *sep;
};

static void
append_topic(gpointer data, gpointer user_data)
{
    struct measurement *m = (struct measurement *) data;
    struct log_data *ldp = (struct log_data *) user_data;

    if (ldp->dst != NULL)
    {
        fprintf(ldp->dst, "%s\"%s\"", ldp->sep, m->topic);
	ldp->sep = ",";
    }

    return;
}

static void
append_value(gpointer data, gpointer user_data)
{
    struct measurement *m = (struct measurement *) data;
    struct log_data *ldp = (struct log_data *) user_data;

    if (ldp->dst != NULL)
    {
        fprintf(ldp->dst, "%s\"%s\"", ldp->sep, m->value);
	ldp->sep = ",";
    }

    return;
}

static void
record_measurement(GList *results, char *pname)
{
    FILE *dst;
    char filename[BUFSIZ];
    struct tm *tv;
    time_t now;
    struct log_data ld;

    time(&now);
    tv = localtime(&now);
    ld.tv = tv;
    snprintf(filename, sizeof(filename), "/var/log/%s-%04d%02d%02d.csv",
        pname, tv->tm_year+1900, tv->tm_mon+1, tv->tm_mday);

    if (access(filename, F_OK) == 0)
        dst = fopen(filename, "a");
    else
    {
        dst = fopen(filename, "w");
	ld.dst = dst;
	ld.sep = "";
	if (dst != NULL)
	    g_list_foreach(results, &append_topic, &ld);
    }

    if (dst == NULL)
        return;

    ld.sep = "";
    g_list_foreach(results, &append_value, &ld);

    fclose(dst);

    return;
}
